/* 
 * This is to write out to the disk in the case that we have
 * essentially a non-functional TiVo.  The idea here is we're
 * going to write out ot the swap partition some log of what
 * happened.  The two cases to get us here generally are
 * a kernel panic, or a watch dog fire.
 *
 * We are going to write out to swap, on boot up, we'll scan
 * the swap before turning it on to see our signature.  If
 * the signature is present, then we'll extract the data
 * and put it in our normal logs.
 *
 * Copyright TiVo (C) 2005
 * Written by Shane Nay (snay@tivo.com)
 *
 */

#include <linux/types.h>
#include <linux/ide.h>
#include <linux/major.h>
#include <linux/delay.h>
#include <linux/ide-deadsystem.h>


#define SWAPFS_MAJOR IDE0_MAJOR
#define SWAPFS_MINOR 8

static int pio_busy_wait(unsigned int usecs, ide_drive_t* drive) {
	int counter = 0;
	while(IN_BYTE(IDE_STATUS_REG) & BUSY_STAT) {
		udelay(1);
		if(counter++ > usecs) { 
			printk("IDE Dead System: permanent busy state\n");
			return -ENODEV;
		}
	}
	return 0;
}

static int pio_wait(ide_drive_t* drive) {
	pio_busy_wait(100000, drive); /* 100 ms */
	if(!OK_STAT(IN_BYTE(IDE_STATUS_REG), READY_STAT,
		    BUSY_STAT|WRERR_STAT /* DF */ |DRQ_STAT|ERR_STAT)) {
		printk("IDE Dead System: drive in bad state 0x%x\n", IN_BYTE(IDE_STATUS_REG));
		return -ENODEV;
	}
	return 0;
}

void inline lba48_address_write(ide_drive_t* drive, unsigned long sector) {
	OUT_BYTE(0 , IDE_NSECTOR_REG);
	OUT_BYTE(1 , IDE_NSECTOR_REG);
	OUT_BYTE(0 , IDE_HCYL_REG); /* high */
	OUT_BYTE(0 , IDE_LCYL_REG); /* mid */
	OUT_BYTE((sector>>24) & 0xff , IDE_SECTOR_REG);/* low */

	OUT_BYTE((sector>>16) & 0xff , IDE_HCYL_REG);  /* high */
	OUT_BYTE((sector>>8 ) & 0xff , IDE_LCYL_REG);  /* mid */
	OUT_BYTE(sector & 0xff     , IDE_SECTOR_REG);  /* low */
	OUT_BYTE(0x40   , IDE_SELECT_REG);
	
	OUT_BYTE(0x34, IDE_COMMAND_REG);
}

void inline lba28_address_write(ide_drive_t* drive, unsigned long sector) {
	OUT_BYTE(1 , IDE_NSECTOR_REG);
	OUT_BYTE(sector & 0xff       , IDE_SECTOR_REG); /* low */
	OUT_BYTE((sector>>8 ) & 0xff , IDE_LCYL_REG);   /* mid */
	OUT_BYTE((sector>>16) & 0xff , IDE_HCYL_REG);   /* high */ 
	OUT_BYTE(((sector>>24) & 0xf) | 0x40  ,	 IDE_SELECT_REG); /* highest */
	
	OUT_BYTE(0x30, IDE_COMMAND_REG);
}

#ifdef CONFIG_TIVO_EIGER

/* This is so broken, S2 is broken, but we treat everything else like its 
 * broken and S2 is cheery.  Fantabulous.  Hrmph.
 */

#define SF_OUT_WORD(data, reg) OUT_WORD(swab16(data), reg)
#else
#define SF_OUT_WORD(data, reg) OUT_WORD(data, reg)
#endif

int pio_write_buffer(ide_drive_t* drive, unsigned long sector,
		     const u8* buf, unsigned int size, u32 *csum) {
	int ret;
	int writtensects=0;
	int i;
	u32 checksum = 0, carry = 0;
	
	while(size > 0) {
		/* Written according to ATA - 6 specification */
		int cnt=0;
		if ((drive->id->cfs_enable_2 & 0x0400)
		    && (drive->addressing)) {
			/* LBA 48 addressing */
			lba48_address_write(drive, sector);
		} else if (drive->select.b.lba) {
			/* LBA 28 addressing */
			lba28_address_write(drive, sector);
		} else {
			printk("IDE Dead System doesn't do CHS\n");
			return -ENODEV;
		}

		udelay(1);

		if((ret = pio_busy_wait(100000, drive)) < 0)
			return ret;

		for(i=(512/2);i;i--) { 
			u16 data;
			/* Unaligned access unwise, makes for funky code */
			if(size >= 2) {
				data = *buf++;
				data |= (*buf++)<<8;
				size -= 2;
			} else if (size == 1){
				data = *buf++;
				size = 0;
			} else {
				data = 0;
			}
			
			SF_OUT_WORD(data, IDE_DATA_REG);
			checksum += carry;
			checksum += data;
			carry = (data > checksum);
			cnt++;
		}
		mdelay(5);

		if((ret = pio_wait(drive)) < 0)
			return ret;
;
		sector++;
		writtensects++;
	}

	checksum += carry;
	
	if(csum != NULL) {
		*csum += checksum;
		if(checksum > *csum)
			*csum = (*csum) + 1;
	}
	
	return writtensects;
}

static int pio_reset(ide_drive_t* drive) {
	int retries=10;

	while(--retries) {
		/* Delays are ATAPI-6 spec times 5 */
		OUT_BYTE(/* nIEN */ 2 | /* SRST */ 4, IDE_CONTROL_REG);
		udelay(50);
		OUT_BYTE(/* nIEN */ 2, IDE_CONTROL_REG);
		mdelay(10);
		if(pio_busy_wait(250000, drive) == 0) { /* Up to 250 ms */
			if(OK_STAT(IN_BYTE(IDE_STATUS_REG), 0,
				    DRQ_STAT|ERR_STAT)) {
				break;
			}
		}
	}

	if(retries == 0) {
		printk("IDE Dead System: drive won't reset, giving up\n");
		return -ENODEV;
	}

	OUT_BYTE(0, IDE_SELECT_REG);
	udelay(1);
	if(pio_wait(drive) < 0) // Device ready?
		return -ENODEV;
	/*
	 * According to the spec, that's all we *have* to do
	 * since we already have the info about the drive that
	 * the kernel queried out.  If this gets buggy on some
	 * drives, we might try adding an IDENTIFY command (0xec
	 * I think)
	 */

	return 0;
}

/* Returns number of sectors written */
int pio_write_buffer_to_swap(const u8* buf, unsigned int size, int secoffset,
	u32 *csum) {
	ide_drive_t* drive= get_info_ptr(MKDEV(SWAPFS_MAJOR, SWAPFS_MINOR));
	struct gendisk *gp= get_gendisk(MKDEV(SWAPFS_MAJOR, 0));
	static int initdrive=1;
	unsigned long tsector;

	if(gp == NULL || drive == NULL)
		return -ENODEV;

	/*
	 * We adjust secoffset to write to the last 1MB of the device,
	 * this is because we don't want to rebuild swap mostly, and
	 * because we can use "lspanic" to grab the last tivo panic.
	 *
	 * We will track which one we've logged by bit-inverting
	 * the checksum.
	 *
	 */

	/* Partition must be at least 2 MB */
	if(gp->part[SWAPFS_MINOR].nr_sects < (2* 1024 * 1024 / 512))
		return -ENODEV;

	/* Offset by 1MB from end of partition */
	secoffset += gp->part[SWAPFS_MINOR].nr_sects - (1 * 1024 * 1024 /512);

	/* Make sure partition is sane size-wise */

	if(gp->part[SWAPFS_MINOR].nr_sects < secoffset)
		return -ENODEV;
	if(gp->part[SWAPFS_MINOR].nr_sects - secoffset < ((size + 511)/512)) {
		return -ENODEV;
	}
	
	if(initdrive) {
                /* Issue software reset, disable ide interrupts */
		if(pio_reset(drive) < 0) { 
			return -ENODEV;
		}
		initdrive=0;
	}

	tsector = gp->part[SWAPFS_MINOR].start_sect + secoffset;

	return pio_write_buffer(drive, tsector, buf, size, csum);

}
