/*
 * Copyright (C) 2001-2003, 2005 TiVo, Inc. All Rights Reserved.
 */

#include <linux/sched.h>
#include <linux/unistd.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/proc_fs.h>
#include <linux/plog.h>
#include <linux/interrupt.h>
#include <linux/notifier.h>
#include <linux/tivoconfig.h>
#include <linux/blkdev.h>
#include <linux/tivo-private.h>
#include <linux/tivo-monotime.h>

#include <asm/uaccess.h>
#include <asm/io.h>

#define PLOG_PR_ALL	0x1
#define PLOG_PR_INTR	0x2
#define PLOG_PR_BH	0x4
#define PLOG_PR_RQ	0x8
#define PLOG_PR_SW	0x10
#define PLOG_PR_SC	0x20
#define PLOG_PR_GEN	0x40
#define PLOG_PR_DRQ	0x80

#define PLOG_DIFF(a,b) ((unsigned long)(a) - (unsigned long)(b))

extern int plog_print_syscall(int num);

extern int mem_get_contigmem_region(unsigned int region, unsigned long *start,
                                        unsigned long *size);

// Anyone can open
static int plog_open(struct inode * inode, struct file * file)
{
    return 0;
}

// close?  Yea, whatever
static int plog_release(struct inode * inode, struct file * file)
{
    return 0;
}

/* Event header:
 *
 *  int eventLen;   // length of event
 *  int eventType;  // predefined type
 *  unsigned long timeStamp;    // CPU instruction clock
 *
 *  int pid;        // For user events only.
 */

// This define controls whether plog is allocated
// as a static array or as a contigmem region.
//#define USE_STATIC_PLOG

#ifndef USE_STATIC_PLOG
static int *processLog = NULL;
static unsigned long processLogSizeBytes = 0;
#define PLOG_LENGTH processLogSizeBytes

// Leave plog disabled until module initialization time
// If you need to debug system bootup with plog, define
// USE_STATIC_PLOG
static int plog_disable = 1;

#else // USE_STATIC_PLOG
static int processLog[32768];
#define PLOG_LENGTH sizeof(processLog)
static int plog_disable = 0;

#endif // !USE_STATIC_PLOG

#define MAX_EVENT   8192
// A 1 in this mask means that the event should not be logged.
static int ptypeMask[MAX_EVENT >> 5];

// The size of a "safety zone" between the write pointer and the read pointer.
// This zone is to mostly insure that when plog_read() tries to read the buffer
// and re-enables interrupts that the place in the buffer that it starts
// reading from doesn't immediately get overwritten by things getting logged.
#define PLOG_SAFETY_ZONE    1024

#define PLOG_MAX_DATASIZE   100

#define PLOG_CONTIGMEM_REGION  0


/* Read/Write pointers are in bytes */
static int plog_wp = 0;
static int plog_rp = 0;

static wait_queue_head_t plog_wait;

static ssize_t plog_read(struct file *file, char *buf, size_t len, loff_t *off)
{
    int error = -EPERM;
    int eventLen;
    int readable;
    int *s;
    int i;
    int *userBuf;

    // The user buffer *must* be word aligned for efficiency (would hate to
    // slow things down significantly as we are trying to profile them).
    error = -EINVAL;
    if (!buf || len < 0) goto out;
    error = 0;
    if (((int) buf) & 0x3) goto out;
    if (!len) goto out;
    error = verify_area(VERIFY_WRITE,buf,len);
    if (error) goto out;

    userBuf = (int *) buf;

    cli();
    if (plog_wp == PLOG_LENGTH) plog_wp = 0;

    error = -ERESTARTSYS;
    if (plog_wp > plog_rp) {
        readable = plog_wp - plog_rp;
    } else {
        readable = plog_wp + (PLOG_LENGTH - plog_rp);
    }
    // Must have 4K available before we let them read anything.
    while (readable < 4096) {
        if (signal_pending(current)) {
            sti();
            goto out;
        }
// printk("sleeping on plog_wait\n");
        interruptible_sleep_on(&plog_wait);
	cli();
        if (plog_wp >= plog_rp) {
            readable = plog_wp - plog_rp;
        } else {
            readable = plog_wp + (PLOG_LENGTH - plog_rp);
        }
    }
    i = 0;
    while (plog_rp != plog_wp && len > 0) {
        eventLen = processLog[plog_rp >> 2];
        if (eventLen == 4) {
            // null event
            plog_rp += 4;
            if (plog_rp == PLOG_LENGTH) plog_rp = 0;
            continue;
        }

        if (len >= eventLen) {
            len -= eventLen;
            s = processLog + (plog_rp >> 2);
            plog_rp += eventLen;
            if (plog_rp == PLOG_LENGTH) plog_rp = 0;
            i += eventLen;

            // I believe that interrupts must be left on for this...
            sti();
            while (eventLen > 0) {
                __put_user(*s, userBuf);
                eventLen -= 4;
                s++;
                userBuf++;
            }
            cli();
        } else break;
    }
    sti();
    error = i;
// printk("Sent out %d/%ld bytes\n", i, len+i);

out:
    return error;
}

static ssize_t plog_write(struct file *file, const char *buf, size_t len,
        loff_t *off)
{
    int error = -EPERM;
    char tempbuf[1024];
    const char *s;
    char *d;

    error = -EINVAL;
    if (!buf || len < 0 || len >= 1023) goto out;
    error = 0;
    if (!len) goto out;
    error = verify_area(VERIFY_READ,buf,len);
    if (error) goto out;

    s = buf;
    d = tempbuf;
    error = len;
    while (len > 0) {
        __get_user(*d, s);
        s++;
        d++;
        len--;
    }
    *d = 0;

    // Parse user's request
    // Request should be in form of "[+-]e1[-e2][,...]"
    s = tempbuf;
    while (*s != 0) {
        int add;
        int start, end;
        int i;

        if (*s == '+') {
            add = 1;
        } else if (*s == '-') {
            add = 0;
        } else {
            error = -EINVAL;
            break;
        }
        s++;
        start = 0;
        while (*s >= '0' && *s <= '9') {
            start = start*10 + (*s - '0');
            s++;
        }
        if (*s == '-') {
            s++;
            end = 0;
            while (*s >= '0' && *s <= '9') {
                end = end*10 + (*s - '0');
                s++;
            }
        } else {
            end = start;
        }
        if (*s != 0 && *s != ',' && *s != '\n') {
            error = -EINVAL;
            break;
        }
        if (*s == ',') s++;
        if (*s == '\n') s++;

        if (start < 0) start = 0;
        if (end >= MAX_EVENT) end = MAX_EVENT - 1;
        if (end >= start) {
            for (i = start; i <= end; i++) {
                if (add) {
                    ptypeMask[i >> 5] &= ~(1 << (i & 0x1f));
                } else {
                    ptypeMask[i >> 5] |= (1 << (i & 0x1f));
                }
            }
        }
    }

out:
    return error;
}

static int plog_assign_region(void)
{
#ifndef USE_STATIC_PLOG
    unsigned long plog_region_start;
    unsigned long plog_size;

    // Only initialize once
    if (processLog != NULL) {
        return 0;
    }

    if (mem_get_contigmem_region(
                PLOG_CONTIGMEM_REGION, 
                &plog_region_start, 
                &plog_size))
    {
        printk("Failed to assign contigmem region for plog\n");
        plog_disable = 1;
        return -EINVAL;
    }
    if (plog_size > 0) {
        processLogSizeBytes = plog_size;
        processLog = (int *)plog_region_start;
        printk("Assigned plog region %d bytes @ %p\n", plog_size, plog_region_start);
    }
#endif
    return 0;
}

/* Write to the process log. */
static void _plLogEvent(int type, int dataLength, void *data)
{
    int *d, *s, *d_start;
    int spaceNeeded;
    unsigned long flags;
    int readable;
    int skipLength;

    // Are we rejecting this type of event?
    if (plog_disable != 0 || (type >= 0 && type < MAX_EVENT &&
            (ptypeMask[type >> 5] & (1 << (type & 0x1f))) != 0)) {
        return;
    }

    /* Data can only be stored in 4 byte increments. */
    dataLength = (dataLength + 3) & ~0x3;

    if (dataLength > PLOG_MAX_DATASIZE) {
        dataLength = PLOG_MAX_DATASIZE;
    }

    /* Allocate space. */
    spaceNeeded = 12 + dataLength;

    /* This code is *not* multithread safe, so we lock it */
    save_flags(flags);
    cli();

    if (plog_wp >= plog_rp) {
        readable = plog_wp - plog_rp;
    } else {
        readable = plog_wp + (PLOG_LENGTH - plog_rp);
    }

    // Compute how much of the buffer we need to auto-consume to leave the
    // SAFETY_ZONE there.
    skipLength = PLOG_SAFETY_ZONE - (PLOG_LENGTH - readable);
    if (PLOG_LENGTH - plog_wp < spaceNeeded) {
        skipLength += PLOG_LENGTH - plog_wp;
    }

    // Advance the read pointer to leave the safety zone large enough.
    while (skipLength > 0) {
        int eventSize = processLog[plog_rp >> 2];

        skipLength -= eventSize;
        plog_rp += eventSize;
        if (plog_rp == PLOG_LENGTH) {
            plog_rp = 0;
        }
    }

    // Advance the write pointer if we cannot fit the event right here.
    if (PLOG_LENGTH - plog_wp < spaceNeeded) {
        while (plog_wp != PLOG_LENGTH) {
            processLog[plog_wp >> 2] = 4;
            plog_wp += 4;
        }
        plog_wp = 0;
    }

    /* Write log entry */
    d_start = d = processLog + (plog_wp >> 2);
    plog_wp += spaceNeeded;
    *d++ = spaceNeeded;
    *d++ = type;
// This seems to be broken for PPC, where it was originally mfspr(SPRN_TBLO)
    *d++ = (int)get_cycles();
    s = data;
    while (dataLength > 0) {
        *d++ = *s++;
        dataLength -= 4;
    }

    restore_flags(flags);

#if 1
    // Flush the event out to DRAM, so it's not stuck in the cache if
    // things just happen to lock up, in which case we might want to
    // look at the process log to find out why...
    dma_cache_wback_inv((unsigned long)d_start, (unsigned long)spaceNeeded);
#endif

    // Wake someone up?
    if (readable >= 4096 && type != PLOG_TYPE_DRQ ) {
        wake_up_interruptible(&plog_wait);
    }
}

int _plog_user(struct descPlogEvent *desc)
{
    int type, data_len;
    void *data;
    int error;
    int buffer[PLOG_MAX_DATASIZE >> 2];
    int *s, *d;
    int copy_len;

    if (get_user(type, &desc->type) ||
        get_user(data_len, &desc->len) ||
        get_user(data, &desc->data)) {
        return -EFAULT;
    }

    buffer[0] = current->pid;
    error = -EINVAL;
    if (data_len < 0) goto out;
    // We are already using 4 bytes for pid, so we reduce data_len
    // accordingly.
    if (data_len > PLOG_MAX_DATASIZE - 4) data_len = PLOG_MAX_DATASIZE - 4;
    if (data_len != 0) {
        if (((int) data) & 0x3) goto out;
        error = verify_area(VERIFY_READ,data,data_len);
        if (error) goto out;
    }

    s = (int *) data;
    d = buffer + 1;
    copy_len = data_len;
    while (copy_len > 0) {
        __get_user(*d, s);
        s++;
        d++;
        copy_len -= 4;
    }

    _plLogEvent(type, data_len+4, buffer);
    error = 0;

out:
    return error;
}

// This is the exported module interface, there is less checking than user
// because the caller is assumed to know what he's doing...
//
int process_log(int type, int data_len, void *data)
{
    int buffer[PLOG_MAX_DATASIZE / sizeof(int)];

    buffer[0] = current->pid;
    if (data_len > PLOG_MAX_DATASIZE - 4) data_len = PLOG_MAX_DATASIZE - 4;
    memcpy(buffer + 1, data, data_len);

    _plLogEvent(type, data_len + 4, buffer);
    return 0;
}

EXPORT_SYMBOL(process_log);

void _plog_swcpu(struct task_struct *from, struct task_struct *to)
{
    struct _pl_swe entry;

    entry.frompid = from->pid;
    entry.topid = to->pid;
    entry.rt_priority = to->rt_priority;
    entry.policy = to->policy;
    _plLogEvent(PLOG_TYPE_SWE, sizeof(entry), &entry);
}

void _plog_swcpu_end(void)
{
    _plLogEvent(PLOG_TYPE_SWL, 0, NULL);
}

void plog_sc_enter(int sc, unsigned long ret_addr) 
{
    struct _pl_sce entry;

    entry.pid = current->pid;
    entry.scnum = sc;
    entry.raddr = ret_addr;
    _plLogEvent(PLOG_TYPE_SCE, sizeof(entry), &entry);
}

void plog_sc_leave(int rv) 
{
    struct _pl_scl entry;

    entry.pid = current->pid;
    entry.rv = rv;
    _plLogEvent(PLOG_TYPE_SCL, sizeof(entry), &entry);
}

// Pass in need_resched instead of pulling it from current, because if we
// wake up from interrupt in the middle of schedule, we may well preempt
// the process that is about the be current, not the one that already is
// (see kernel/sched.c)
//
void _plog_rq_add(struct task_struct *tp, long need_resched) 
{
    struct _pl_rq entry;

    entry.pid = (short)tp->pid;
    entry.fromintr = in_interrupt();
    entry.preempt = need_resched ? 1 : 0;
    entry.priority = (short)tp->nice;
    entry.rt_priority = (short)tp->rt_priority;
    entry.policy = (short)tp->policy;
    _plLogEvent(PLOG_TYPE_RQ, sizeof(entry), &entry);
}

void _plog_rq_del(struct task_struct *tp) 
{
    struct _pl_drq entry;

    entry.pid = (short)tp->pid;
    entry.priority = (short)tp->nice;
    entry.rt_priority = (short)tp->rt_priority;
    entry.policy = (short)tp->policy;
    _plLogEvent(PLOG_TYPE_DRQ, sizeof(entry), &entry);
}

void _plog_setscheduler(int pid, int sched_pid, short priority, short policy) {
    int buffer[3];

    buffer[0]= pid;
    buffer[1]= sched_pid;
    buffer[2]= (priority << 16) | policy;
    _plLogEvent(PLOG_TYPE_SETSCHED, sizeof(buffer), buffer);
}

void _plog_intr_enter(int num, unsigned long pc)
{
    struct _pl_intre entry;

    entry.irq = num;
    entry.pc = pc;
    _plLogEvent(PLOG_TYPE_INTRE, sizeof(entry), &entry);
}

void _plog_intr_leave(int num)
{
    _plLogEvent(PLOG_TYPE_INTRL, sizeof(num), &num);
}

void _plog_bh_enter(int num)
{
    _plLogEvent(PLOG_TYPE_BHE, sizeof(num), &num);
}

void _plog_bh_leave(int num)
{
    _plLogEvent(PLOG_TYPE_BHL, sizeof(num), &num);
}

void _plog_filemap_page_swap(int pid, int physPage, int virtPage)
{
    int buffer[3];

    buffer[0] = pid;
    buffer[1] = physPage;
    buffer[2] = virtPage;
    _plLogEvent(PLOG_TYPE_FMS, sizeof(buffer), buffer);
}

void _plog_generic_page_swap(int pid, int physPage, int virtPage)
{
    int buffer[3];

    buffer[0] = pid;
    buffer[1] = physPage;
    buffer[2] = virtPage;
    _plLogEvent(PLOG_TYPE_GPS, sizeof(buffer), buffer);
}

void _plog_reswap(int pid, int physPage, int virtPage)
{
    int buffer[3];

    buffer[0] = pid;
    buffer[1] = physPage;
    buffer[2] = virtPage;
    _plLogEvent(PLOG_TYPE_RESWAP, sizeof(buffer), buffer);
}

void _plog_pageout(int pid, int physPage, int virtPage)
{
    int buffer[3];

    buffer[0] = pid;
    buffer[1] = physPage;
    buffer[2] = virtPage;
    _plLogEvent(PLOG_TYPE_PAGEOUT, sizeof(buffer), buffer);
}

void _plog_diskdispatch(unsigned long start, unsigned long sectors)
{
    int buffer[2];

    buffer[0] = (int)start;
    buffer[1] = (int)sectors;

    _plLogEvent(PLOG_TYPE_DISK_DISPATCH, sizeof(buffer), buffer);
}

void _plog_diskstart(struct request *rq)
{
    int buffer[1];

    if (rq->tivo == NULL) {
        buffer[0] = 0;      // type 0: non-TiVo request
    } else if (tivo_monotonic_is_null(&rq->tivo->deadline)) {
        buffer[0] = 2;      // type 2: TiVo non-RT request
    } else {
        buffer[0] = 1;      // type 1: TiVo RT request
    }

    _plLogEvent(PLOG_TYPE_DISK_START, sizeof(buffer), buffer);
}

void _plog_diskstop(struct request *rq)
{
    int buffer[3];
    struct idetivo_request *trq = rq->tivo;

    buffer[0] = rq->nr_sectors;
    if (trq != NULL && rq->nr_sectors > trq->max_nrsectors) {
        buffer[0] = trq->max_nrsectors;
    }

    buffer[1] = rq->cmd == READ ? 0 : 1;

    if (trq == NULL || tivo_monotonic_is_null(&trq->deadline)) {
        buffer[2] = 0;  // no deadline
    } else {
        monotonic_t now_time;
        s64 diff_count;

        tivo_monotonic_get_current(&now_time);

        diff_count = (s64)(tivo_monotonic_retrieve(&trq->deadline) -
                           tivo_monotonic_retrieve(&now_time));
        // clamp to min/max s32
        if (diff_count > (s64)0x000000007fffffffLL) {
            buffer[2] = 0x7fffffff;
        } else if (diff_count < (s64)(0xffffffff80000000LL)) {
            buffer[2] = 0x80000000;
        } else {
            buffer[2] = (int)diff_count;
        }
    }

    _plLogEvent(PLOG_TYPE_DISK_STOP, sizeof(buffer), buffer);
}

void _plog_generic_event0(int event)
{
    _plLogEvent(event, 0, NULL);
}

void _plog_generic_event(int event, int somedata)
{
    int buffer[2];
    if (current == NULL) {
        buffer[0] = -1;
    } else {
        buffer[0] = current->pid;
    }
    buffer[1] = somedata;
    _plLogEvent(event, sizeof(buffer), buffer);
}

void _plog_generic_event2(int event, int data1, int data2)
{
    int buffer[3];
    if (current == NULL) {
        buffer[0] = -1;
    } else {
        buffer[0] = current->pid;
    }
    buffer[1] = data1;
    buffer[2] = data2;
    _plLogEvent(event, sizeof(buffer), buffer);
}

void _plog_generic_event3(int event, int data1, int data2, int data3)
{
    int buffer[4];
    if (current == NULL) {
        buffer[0] = -1;
    } else {
        buffer[0] = current->pid;
    }
    buffer[1] = data1;
    buffer[2] = data2;
    buffer[3] = data3;
    _plLogEvent(event, sizeof(buffer), buffer);
}

void _plog_ioctl(unsigned int fd, unsigned int cmd, unsigned long arg)
{
    struct _pl_ioctl entry;

    entry.pid = current->pid;
    entry.fd = fd;
    entry.cmd = cmd;
    entry.arg = arg;
    _plLogEvent(PLOG_TYPE_IOCTL, sizeof(entry), &entry);
}

static int eventWanted(int *event, int flags, int pid)
{
    switch (event[1]) {
      case PLOG_TYPE_SWE:
        {
            struct _pl_swe *logEntry;

            logEntry = (struct _pl_swe *) (event + 3);
            if (pid && (logEntry->frompid == pid || logEntry->topid == pid)) {
                return 1;
            }
            if (!(flags & PLOG_PR_SW)) return 0;
        }
        break;

      case PLOG_TYPE_SWL:
        if (!(flags & PLOG_PR_SW)) return 0;
        break;

      case PLOG_TYPE_SCE:
        {
            struct _pl_sce *logEntry;

            logEntry = (struct _pl_sce *) (event + 3);
            if (pid && logEntry->pid == pid) return 1;
            if (!(flags & PLOG_PR_SC)) return 0;
        }
        break;

      case PLOG_TYPE_SCL:
        {
            struct _pl_scl *logEntry;

            logEntry = (struct _pl_scl *) (event + 3);
            if (pid && logEntry->pid == pid) return 1;
            if (!(flags & PLOG_PR_SC)) return 0;
        }
        break;

      case PLOG_TYPE_IOCTL:
        {
            struct _pl_ioctl *logEntry;

            logEntry = (struct _pl_ioctl *) (event + 3);
            if (pid && logEntry->pid == pid) return 1;
            if (!(flags & PLOG_PR_SC)) return 0;
        }
        break;

      case PLOG_TYPE_BHE:
      case PLOG_TYPE_BHL:
        if (!(flags & PLOG_PR_BH)) return 0;
        break;

      case PLOG_TYPE_RQ:
        {
            struct _pl_rq *logEntry;

            logEntry = (struct _pl_rq *) (event + 3);
            if (pid && logEntry->pid == pid) return 1;
            if (!(flags & PLOG_PR_RQ)) return 0;
        }
        break;

    case PLOG_TYPE_DRQ:
	{
	    struct _pl_drq *logEntry;

	    logEntry = (struct _pl_drq *) (event + 3);
	    if (pid && logEntry->pid == pid) return 1;
	    if (!(flags & PLOG_PR_DRQ)) return 0;
	}
	break;

      case PLOG_TYPE_INTRE:
      case PLOG_TYPE_INTRL:
        if (!(flags & PLOG_PR_INTR)) return 0;
        break;

      default:
        break;
    }
    return 1;
}

void plog_dump(int max, int swflags) 
{
    int plog_iter;
    int count;
    int eventSize;
    int first;
    unsigned long lastTime = 0;
    unsigned long firstTime = 0;
    int eventType;
    int nextEventType;
    unsigned long thisTime;
    unsigned long nextTime = 0;
    void *data;
    void *nextData = NULL;
    unsigned long flags;
    int pid;
    int nextUsed;

    save_flags(flags);
    cli();

    if (max == 0) max = 200;
    if (plog_wp == PLOG_LENGTH) plog_wp = 0;
    if (max < 0) {
        pid = -max;
        max = 400;
    } else pid = 0;

    plog_iter = plog_rp;
    count = 0;
    while (plog_iter != plog_wp) {
        eventSize = processLog[plog_iter >> 2];
        if (eventSize != 4 && 
                eventWanted(processLog + (plog_iter >> 2), swflags, pid)) {
            count++;
        }
        plog_iter += eventSize;
        if (plog_iter == PLOG_LENGTH) plog_iter = 0;
    }

    printk("%d relevant events captured\n", count);
    plog_iter = plog_rp;
    if (count > max) {
        count -= max;
        printk("Skipping %d events\n", count);
        while (count > 0) {
            eventSize = processLog[plog_iter >> 2];
            if (eventSize != 4 && 
                    eventWanted(processLog + (plog_iter >> 2), swflags, pid)) {
                count--;
            }
            plog_iter += eventSize;
            if (plog_iter == PLOG_LENGTH) plog_iter = 0;
        }
    }

    first = 1;
    while (plog_iter != plog_wp) {
        eventSize = processLog[plog_iter >> 2];
        if (eventSize == 4 || 
                !eventWanted(processLog + (plog_iter >> 2), swflags, pid)) {
            plog_iter += eventSize;
            continue;
        }
        eventType = processLog[1 + (plog_iter >> 2)];
        thisTime = processLog[2 + (plog_iter >> 2)];
        if (first) {
            first = 0;
            lastTime = thisTime;
            firstTime = thisTime;
        }
        data = (void *) (processLog + 3 + (plog_iter >> 2));
        plog_iter += eventSize;
        if (plog_iter == PLOG_LENGTH) plog_iter = 0;

        /* Find next event for event start/end merging */
        while (plog_iter != plog_wp && processLog[(plog_iter>>2)] == 4) {
            plog_iter += 4;
            if (plog_iter == PLOG_LENGTH) plog_iter = 0;
        }
        if (plog_iter != plog_wp) {
            nextEventType = processLog[1 + (plog_iter >> 2)];
            nextTime = processLog[2 + (plog_iter >> 2)];
            nextData = (void *) (processLog + 3 + (plog_iter >> 2));
        } else {
            nextEventType = 0;
        }
        nextUsed = 0;
                
        printk("%-10ld | ^%-6lu ", PLOG_DIFF(thisTime, firstTime),
                PLOG_DIFF(thisTime, lastTime));
        lastTime = thisTime;
        switch (eventType) {
          case PLOG_TYPE_SWE:
            if (nextEventType == PLOG_TYPE_SWL) {
                printk("Switch: %d->%d (PRI:%d POL%d %lu cyc)\n",
                        ((struct _pl_swe *) data)->frompid,
			((struct _pl_swe *) data)->topid,
                        ((struct _pl_swe *) data)->rt_priority,
			((struct _pl_swe *) data)->policy,
                        PLOG_DIFF(nextTime, thisTime));
                nextUsed = 1;
            } else {
                printk("Switch Start: %d->%d (PRI:%d POL:%d)\n", 
                        ((struct _pl_swe *) data)->frompid,
			((struct _pl_swe *) data)->topid,
                        ((struct _pl_swe *) data)->rt_priority,
			((struct _pl_swe *) data)->policy);
            }
            break;

          case PLOG_TYPE_SWL:
            printk("Switch End\n");
            break;

          case PLOG_TYPE_SCE:
            {
                int callnum = ((struct _pl_sce *) data)->scnum;

                printk("syscall%s: pid=%d: ", 
                        nextEventType == PLOG_TYPE_SCL ? "" : " start",
                        ((struct _pl_sce *) data)->pid);
                if (plog_print_syscall(callnum) != 0) {
                    printk("syscall(%d)", callnum);
                }
                printk(" (ra=0x%lx)", ((struct _pl_sce *) data)->raddr);
                if (nextEventType == PLOG_TYPE_SCL) {
                    printk(" == %d (%lu cyc)",
                            ((struct _pl_scl *) nextData)->rv,
                            PLOG_DIFF(nextTime, thisTime));
                    nextUsed = 1;
                }
                printk("\n");
            }
            break;

          case PLOG_TYPE_SCL:
            printk("syscall finish: pid=%d, rv=%d\n",
                    ((struct _pl_scl *) data)->pid,
                    ((struct _pl_scl *) data)->rv);
            break;

          case PLOG_TYPE_IOCTL:
                printk("ioctl: pid=%d, fd=%d, cmd=%d, arg=0x%lx\n", 
                        ((struct _pl_ioctl *) data)->pid,
                        ((struct _pl_ioctl *) data)->fd,
                        ((struct _pl_ioctl *) data)->cmd,
                        ((struct _pl_ioctl *) data)->arg);
            break;

          case PLOG_TYPE_RQ:
            printk("pid %d runnable pri=%d rt=%d policy=%d (%sfromintr, %spreempt)\n",
                    ((struct _pl_rq *) data)->pid,
		    ((struct _pl_rq *) data)->priority,
		    ((struct _pl_rq *) data)->rt_priority,
		    ((struct _pl_rq *) data)->policy,
                    ((struct _pl_rq *) data)->fromintr ? "" : "!",
                    ((struct _pl_rq *) data)->preempt ? "" : "!");
            break;
	case PLOG_TYPE_DRQ:
	    printk("pid %d unrunnable pri=%d rt=%d policy=%d\n",
		   ((struct _pl_drq *) data)->pid,
		   ((struct _pl_drq *) data)->priority,
		   ((struct _pl_drq *) data)->rt_priority,
		   ((struct _pl_drq *) data)->policy);
	    break;

          case PLOG_TYPE_INTRE:
            if (nextEventType == PLOG_TYPE_INTRL) {
                nextUsed = 1;
                printk("interrupt: inum=%d (pc=0x%lx, %lu cyc)\n",
                        ((struct _pl_intre *) data)->irq,
                        ((struct _pl_intre *) data)->pc,
                        PLOG_DIFF(nextTime, thisTime));
            } else {
                printk("interrupt start: inum=%d (pc=0x%lx)\n",
                        ((struct _pl_intre *) data)->irq,
                        ((struct _pl_intre *) data)->pc);
            }
            break;

          case PLOG_TYPE_INTRL:
            printk("interrupt end: inum=%d\n",
                    *(int *) data);
            break;

          case PLOG_TYPE_BHE:
            if (nextEventType == PLOG_TYPE_BHL) {
                nextUsed = 1;
                printk("Bottom half: %d (%lu cyc)\n",
                        *(int *) data,
                        PLOG_DIFF(nextTime, thisTime));
            } else {
                printk("Bottom half start: %d\n",
                        *(int *) data);
            }
            break;

          case PLOG_TYPE_BHL:
            printk("Bottom half end: %d\n",
                    *(int *) data);
            break;

          case PLOG_TYPE_SPS:
            printk("pid %d: Shared page at 0x%08x swapped out\n",
                    ((int *) data)[0],
                    ((int *) data)[1]);
            break;

          case PLOG_TYPE_FMS:
            printk("pid %d: Data page ot 0x%08x (vaddr=0x%08x) swapped to backing file\n",
                    ((int *) data)[0],
                    ((int *) data)[1],
                    ((int *) data)[2]);
            break;

          case PLOG_TYPE_GPS:
            printk("pid %d: Data page at 0x%08x (vaddr=0x%08x) swapped out\n",
                    ((int *) data)[0],
                    ((int *) data)[1],
                    ((int *) data)[2]);
            break;

          case PLOG_TYPE_RESWAP:
            printk("pid %d: Data page at 0x%08x (vaddr=0x%08x) tossed out\n",
                    ((int *) data)[0],
                    ((int *) data)[1],
                    ((int *) data)[2]);
            break;

          case PLOG_TYPE_PAGEOUT:
            printk("pid %d: Memory at 0x%08x (vaddr=0x%08x) tossed out\n",
                    ((int *) data)[0],
                    ((int *) data)[1],
                    ((int *) data)[2]);
            break;

          case PLOG_TYPE_PAGEIN:
            if (nextEventType == PLOG_TYPE_PAGEIN_DONE) {
                nextUsed = 1;
                printk("pid %d: Memory at vaddr 0x%08x paged in (%lu cyc)\n",
                        ((int *) data)[0],
                        ((int *) data)[1],
                        PLOG_DIFF(nextTime, thisTime));
            } else {
                printk("pid %d: Memory at vaddr 0x%08x page in start\n",
                        ((int *) data)[0],
                        ((int *) data)[1]);
            }
            break;

          case PLOG_TYPE_PAGEIN_DONE:
            printk("pid %d: Memory at vaddr 0x%08x page in completed\n",
                    ((int *) data)[0],
                    ((int *) data)[1]);
            break;

          case PLOG_TYPE_WP_PAGE:
            if (nextEventType == PLOG_TYPE_WP_PAGE_DONE) {
                nextUsed = 1;
                printk("pid %d: Memory at vaddr 0x%08x made writable (%lu cyc)\n",
                        ((int *) data)[0],
                        ((int *) data)[1],
                        PLOG_DIFF(nextTime, thisTime));
            } else {
                printk("pid %d: Memory at vaddr 0x%08x needs copy-on-write\n",
                        ((int *) data)[0],
                        ((int *) data)[1]);
            }
            break;

          case PLOG_TYPE_WP_PAGE_DONE:
            printk("pid %d: Memory at vaddr 0x%08x copy-on-write complete\n",
                    ((int *) data)[0],
                    ((int *) data)[1]);
            break;

          case PLOG_TYPE_DEMAND_ZERO:
            printk("pid %d: Memory at vaddr 0x%08x demand zeroed\n",
                    ((int *) data)[0],
                    ((int *) data)[1]);
            break;

          case PLOG_TYPE_PAGE_DIRTY:
            printk("pid %d: PTE 0x%08x being marked dirty\n",
                    ((int *) data)[0],
                    ((int *) data)[1]);
            break;

          case PLOG_TYPE_NO_SWAP_PAGE:
            printk("pid %d: No swap page available to swap out pid %d, vaddr 0x%08x\n",
                    ((int *) data)[0],
                    ((int *) data)[1],
                    ((int *) data)[2]);
            break;

          case PLOG_TYPE_SHARED_PAGE_IGNORED:
            printk("pid %d: Shared page for pid %d, paddr 0x%08x multiply referenced, not paged\n",
                    ((int *) data)[0],
                    ((int *) data)[1],
                    ((int *) data)[2]);
            break;

          case PLOG_TYPE_DIRTY_MULTIPLE_NO_SWAP:
            printk("pid %d: Page for pid %d, vaddr 0x%08x not swapped out -- it's being DMAed\n",
                    ((int *) data)[0],
                    ((int *) data)[1],
                    ((int *) data)[2]);
            break;

          case PLOG_TYPE_EXIT:
            printk("pid %d exiting with code 0x%x\n",
                    ((int *) data)[0],
                    ((int *) data)[1]);
            break;

          default:
            printk("Event %d: pid=%d, Len=%d",
                    eventType, *(int *) data, eventSize-16);
            if (eventSize > 16) {
                int *idata = ((int *) data) + 1;
                int chlen = eventSize - 16;
                printk (" [");
                while (chlen > 0) {
                    printk("%d", *idata);
                    idata++;
                    chlen -= 4;
                    if (chlen > 0) printk(" ");
                }
                printk ("]\n");
            } else {
                printk ("\n");
            }
        }
        /* If we already consumed the next event, eat it */
        if (nextUsed) {
            eventSize = processLog[plog_iter >> 2];
            plog_iter += eventSize;
            if (plog_iter == PLOG_LENGTH) plog_iter = 0;
        }
    }

    restore_flags(flags);
}

EXPORT_SYMBOL(plog_dump);

/* Support for /proc/plog */
static struct file_operations proc_plog_operations = {
	read:		plog_read,
	write:		plog_write,
	open:		plog_open,
	release:	plog_release,
};

// We usually do a 1 second delay on panic, then reboot.  If ints are
// enabled at the time, that can fill up the process log with interrupt
// events, when we're probably more interested in what caused the panic...
//
static int plog_panic_event(struct notifier_block *self,
			    unsigned long event,
			    void *ptr)
{
    plog_disable = 1;
    return NOTIFY_DONE;
}

/* DO NOT USE THIS FUNCTION unless it's wednesday.
 * No, but seriously, it totally destroys plog, but it
 * gets events exactly as they happened at the insertion
 * point.  Which is what we need for panic logger, but
 * shouldn't be used anywhere else.
 */

void plog_destructive_liforead(int *buf, int entries)
{
    /* Figure out where we should start */
    plog_disable = 1;
    if(processLog != NULL) {
	int plp=plog_wp - (entries<<2);
	if(plp < 0) {
	    plp+=PLOG_LENGTH;
	}
	plog_wp=plp;
	while(entries--) {
	    *buf++=processLog[plp>>2];
	    plp += 4;
	    if(plp >= PLOG_LENGTH) {
		plp=0;
	    }
	}
    }
}

static struct notifier_block plog_panic_block = {
    plog_panic_event,
    NULL,
    0
};

static int __init plog_init(void)
{
    struct proc_dir_entry *entry;

    // Assign the contigmem region for plog
    if (plog_assign_region()) {
        return -ENOMEM;
    }

    plog_disable = 0;

    init_waitqueue_head(&plog_wait);

    entry = create_proc_entry("plog", S_IFREG | S_IRUSR | S_IWUSR, 0);
    if (entry) {
        entry->nlink = 1;
        entry->proc_fops = &proc_plog_operations;

    }

    notifier_chain_register(&panic_notifier_list, &plog_panic_block);

    return 0;
}

module_init(plog_init);


